from collections import Counter
from flask import session
import cqi
import math
import random
from app import db, socketio
from app.decorators import socketio_login_required
from app.models import Corpus
from . import NAMESPACE as ns
from .utils import cqi_over_socketio, lookups_by_cpos


@socketio.on('cqi.corpora.corpus.drop', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_drop(cqi_client: cqi.CQiClient, corpus_name: str):
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    cqi_status = cqi_corpus.drop()
    payload = {'code': cqi_status.code,
               'msg': cqi_status.__class__.__name__}
    return {'code': 200, 'msg': 'OK', 'payload': payload}


@socketio.on('cqi.corpora.corpus.query', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_query(cqi_client: cqi.CQiClient, corpus_name: str, subcorpus_name: str, query: str):  # noqa
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    cqi_status = cqi_corpus.query(subcorpus_name, query)
    payload = {'code': cqi_status.code,
               'msg': cqi_status.__class__.__name__}
    return {'code': 200, 'msg': 'OK', 'payload': payload}


###############################################################################
# nopaque specific CQi extensions                                             #
###############################################################################
@socketio.on('cqi.corpora.corpus.update_db', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_update_db(cqi_client: cqi.CQiClient, corpus_name: str):
    corpus = Corpus.query.get(session['d']['corpus_id'])
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    corpus.num_tokens = cqi_corpus.size
    db.session.commit()


@socketio.on('cqi.corpora.corpus.get_visualization_data', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_get_visualization_data(cqi_client: cqi.CQiClient, corpus_name: str):
    corpus = cqi_client.corpora.get(corpus_name)
    payload = {
        'corpus': {
            'bounds': [0, corpus.size - 1],
            'counts': {},
            'freqs': {}
        },
        'p_attrs': {},
        's_attrs': {},
        'values': {'p_attrs': {}, 's_attrs': {}}
    }
    for p_attr in corpus.positional_attributes.list():
        payload['corpus']['freqs'][p_attr.name] = dict(
            zip(
                range(0, p_attr.lexicon_size),
                p_attr.freqs_by_ids(list(range(0, p_attr.lexicon_size)))
            )
        )
        payload['p_attrs'][p_attr.name] = dict(
            zip(
                range(0, corpus.size),
                p_attr.ids_by_cpos(list(range(0, corpus.size)))
            )
        )
        payload['values']['p_attrs'][p_attr.name] = dict(
            zip(
                range(0, p_attr.lexicon_size),
                p_attr.values_by_ids(list(range(0, p_attr.lexicon_size)))
            )
        )
    s = corpus.structural_attributes.get('s')
    ent = corpus.structural_attributes.get('ent')
    for s_attr in corpus.structural_attributes.list():
        if s_attr.has_values:
            continue
        payload['corpus']['counts'][s_attr.name] = s_attr.size
        payload['s_attrs'][s_attr.name] = {'lexicon': {}, 'values': []}
        for id in range(0, s_attr.size):
            payload['s_attrs'][s_attr.name]['lexicon'][id] = {}
            if s_attr.name != 'text':
                continue
            lbound, rbound = s_attr.cpos_by_id(id)
            payload['s_attrs'][s_attr.name]['lexicon'][id]['bounds'] = [lbound, rbound]
            cpos_range = range(lbound, rbound + 1)
            payload['s_attrs'][s_attr.name]['lexicon'][id]['counts'] = {}
            payload['s_attrs'][s_attr.name]['lexicon'][id]['counts']['s'] = len({x for x in s.ids_by_cpos(list(cpos_range)) if x != -1})
            payload['s_attrs'][s_attr.name]['lexicon'][id]['counts']['ent'] = len({x for x in ent.ids_by_cpos(list(cpos_range)) if x != -1})
            payload['s_attrs'][s_attr.name]['lexicon'][id]['counts']['token'] = rbound - lbound + 1
            payload['s_attrs'][s_attr.name]['lexicon'][id]['freqs'] = {}
            for p_attr in corpus.positional_attributes.list():
                payload['s_attrs'][s_attr.name]['lexicon'][id]['freqs'][p_attr.name] = dict(Counter(p_attr.ids_by_cpos(list(cpos_range))))
    # for s_attr in s_attrs:
    #     payload['s_attrs'][s_attr.name] = {'lexicon': {}, 'values': []}
    #     payload['values']['s_attrs'][s_attr.name] = {}
    #     for id in range(0, s_attr.size):
    #         payload['s_attrs'][s_attr.name]['lexicon'][id] = {}
    #         if s_attr.name != 'text':
    #             continue
    #         lbound, rbound = s_attr.cpos_by_id(id)
    #         cpos_range = range(lbound, rbound + 1)
    #         # s_ids
    #         payload['s_attrs'][s_attr.name]['lexicon'][id]['bounds'] = [lbound, rbound]
    #         payload['s_attrs'][s_attr.name]['lexicon'][id]['counts'] = {}
    #         payload['s_attrs'][s_attr.name]['lexicon'][id]['counts']['token'] = rbound - lbound + 1
    #         payload['s_attrs'][s_attr.name]['lexicon'][id]['freqs'] = {
    #             p_attr.name: dict(Counter(p_attr.ids_by_cpos(list(cpos_range))))
    #             for p_attr in p_attrs
    #         }
    #     for sub_attr in corpus.structural_attributes.list(filters={'part_of': s_attr}):
    #         payload['s_attrs'][s_attr.name]['values'].append(sub_attr.name[(len(s_attr.name) + 1):])
    #         payload['values']['s_attrs'][s_attr.name][sub_attr.name[(len(s_attr.name) + 1):]] = dict(
    #             zip(
    #                 range(0, sub_attr.size),
    #                 sub_attr.values_by_ids(list(range(0, sub_attr.size)))
    #             )
    #         )
    # text = corpus.structural_attributes.get('text')

    # text = corpus.structural_attributes.get('text')
    # text_value_names = []
    # text_values = []
    # for text_sub_attr in corpus.structural_attributes.list(filters={'part_of': text}):
    #     text_value_names.append(text_sub_attr.name[(len(text.name) + 1):])
    #     text_values.append(text_sub_attr.values_by_ids(list(range(0, text.size))))
    # s = corpus.structural_attributes.get('s')
    # ent = corpus.structural_attributes.get('ent')
    # ent_value_names = []
    # ent_values = []
    # for ent_sub_attr in corpus.structural_attributes.list(filters={'part_of': ent}):
    #     ent_value_names.append(ent_sub_attr.name[(len(ent.name) + 1):])
    #     ent_values.append(ent_sub_attr.values_by_ids(list(range(0, ent.size))))
    # word = corpus.positional_attributes.get('word')
    # lemma = corpus.positional_attributes.get('lemma')
    # pos = corpus.positional_attributes.get('pos')
    # simple_pos = corpus.positional_attributes.get('simple_pos')
    # payload = {}

    # payload['text'] = {'lexicon': {}, 'values': None}
    # for text_id in range(0, text.size):
    #     text_lbound, text_rbound = text.cpos_by_id(text_id)
    #     text_cpos_range = range(text_lbound, text_rbound + 1)
    #     text_s_ids = s.ids_by_cpos(list(text_cpos_range))
    #     text_ent_ids = ent.ids_by_cpos(list(text_cpos_range))
    #     payload['text']['lexicon'][text_id] = {
    #         'bounds': [text_lbound, text_rbound],
    #         'counts': {
    #             's': len([x for x in text_s_ids if x != -1]),
    #             'ent': len([x for x in text_ent_ids if x != -1]),
    #             'token': text_rbound - text_lbound + 1
    #         },
    #         'freqs': {
    #             'word': dict(
    #                 Counter(word.ids_by_cpos(list(text_cpos_range)))
    #             ),
    #             'lemma': dict(
    #                 Counter(lemma.ids_by_cpos(list(text_cpos_range)))
    #             ),
    #             'pos': dict(
    #                 Counter(pos.ids_by_cpos(list(text_cpos_range)))
    #             ),
    #             'simple_pos': dict(
    #                 Counter(simple_pos.ids_by_cpos(list(text_cpos_range)))
    #             )
    #         }
    #     }
    # payload['text']['values'] = text_value_names
    # payload['s'] = {'lexicon': {}, 'values': None}
    # for s_id in range(0, s.size):
    #     payload['s']['lexicon'][s_id] = {
    #         # 'bounds': s.cpos_by_id(s_id)
    #     }
    # payload['s']['values'] = [
    #     sub_attr.name[(len(s.name) + 1):]
    #     for sub_attr in corpus.structural_attributes.list(filters={'part_of': s})
    # ]
    # payload['ent'] = {'lexicon': {}, 'values': None}
    # for ent_id in range(0, ent.size):
    #     payload['ent']['lexicon'][ent_id] = {
    #         # 'bounds': ent.cpos_by_id(ent_id)
    #     }
    # payload['ent']['values'] = ent_value_names
    # payload['lookups'] = {
    #     'corpus': {},
    #     'text': {
    #         text_id: {
    #             text_value_name: text_values[text_value_name_idx][text_id_idx]
    #             for text_value_name_idx, text_value_name in enumerate(text_value_names)
    #         } for text_id_idx, text_id in enumerate(range(0, text.size))
    #     },
    #     's': {},
    #     'ent': {
    #         ent_id: {
    #             ent_value_name: ent_values[ent_value_name_idx][ent_id_idx]
    #             for ent_value_name_idx, ent_value_name in enumerate(ent_value_names)
    #         } for ent_id_idx, ent_id in enumerate(range(0, ent.size))
    #     },
    #     'word': dict(
    #         zip(
    #             range(0, word.lexicon_size),
    #             word.values_by_ids(list(range(0, word.lexicon_size)))
    #         )
    #     ),
    #     'lemma': dict(
    #         zip(
    #             range(0, lemma.lexicon_size),
    #             lemma.values_by_ids(list(range(0, lemma.lexicon_size)))
    #         )
    #     ),
    #     'pos': dict(
    #         zip(
    #             range(0, pos.lexicon_size),
    #             pos.values_by_ids(list(range(0, pos.lexicon_size)))
    #         )
    #     ),
    #     'simple_pos': dict(
    #         zip(
    #             range(0, simple_pos.lexicon_size),
    #             simple_pos.values_by_ids(list(range(0, simple_pos.lexicon_size)))
    #         )
    #     )
    # }
    # print(payload)
    return {'code': 200, 'msg': 'OK', 'payload': payload}


@socketio.on('cqi.corpora.corpus.paginate', namespace=ns)
@socketio_login_required
@cqi_over_socketio
def cqi_corpora_corpus_paginate(cqi_client: cqi.CQiClient, corpus_name: str, page: int = 1, per_page: int = 20):  # noqa
    cqi_corpus = cqi_client.corpora.get(corpus_name)
    # Sanity checks
    if (
        per_page < 1
        or page < 1
        or (
            cqi_corpus.size > 0
            and page > math.ceil(cqi_corpus.size / per_page)
        )
    ):
        return {'code': 416, 'msg': 'Range Not Satisfiable'}
    first_cpos = (page - 1) * per_page
    last_cpos = min(cqi_corpus.size, first_cpos + per_page)
    cpos_list = [*range(first_cpos, last_cpos)]
    lookups = lookups_by_cpos(cqi_corpus, cpos_list)
    payload = {}
    # the items for the current page
    payload['items'] = [cpos_list]
    # the lookups for the items
    payload['lookups'] = lookups
    # the total number of items matching the query
    payload['total'] = cqi_corpus.size
    # the number of items to be displayed on a page.
    payload['per_page'] = per_page
    # The total number of pages
    payload['pages'] = math.ceil(payload['total'] / payload['per_page'])
    # the current page number (1 indexed)
    payload['page'] = page if payload['pages'] > 0 else None
    # True if a previous page exists
    payload['has_prev'] = payload['page'] > 1 if payload['page'] else False
    # True if a next page exists.
    payload['has_next'] = payload['page'] < payload['pages'] if payload['page'] else False  # noqa
    # Number of the previous page.
    payload['prev_num'] = payload['page'] - 1 if payload['has_prev'] else None
    # Number of the next page
    payload['next_num'] = payload['page'] + 1 if payload['has_next'] else None
    return {'code': 200, 'msg': 'OK', 'payload': payload}
